import numpy as np


# ESTIMATORS
################################################################################

def mean(sample):
    return sum(sample) / len(sample)


def var(sample):
    xbar = mean(sample)
    sumsqdevs = sum([(xi-xbar)**2 for xi in sample])
    return sumsqdevs / (len(sample)-1)


def std(sample):
    s2 = var(sample)
    return np.sqrt(s2)


def dmeans(xsample, ysample):
    dhat = mean(xsample) - mean(ysample)
    return dhat




# DESCRIPTIVE STATISTICS
################################################################################

def median(values):
    n = len(values)
    svalues = sorted(values)
    if n % 2 == 1:            # Case A: n is odd
        mid = n // 2
        return svalues[mid]
    else:                     # Case B: n is even
        j = n // 2
        return 0.5*svalues[j-1] + 0.5*svalues[j]


def quantile(values, q):
    svalues = sorted(values)
    p = q * (len(values)-1)
    i = int(p)
    g = p - int(p)
    return (1-g)*svalues[i] + g*svalues[i+1]



import matplotlib.pyplot as plt
import numpy as np
import seaborn as sns



# Discrete random variables
################################################################################

def plot_pmf(rv, xlims=None, ylims=None, rv_name="X", ax=None, title=None, label=None):
    """
    Plot the pmf of the discrete random variable `rv` over the `xlims`.
    """
    # Setup figure and axes
    if ax is None:
        fig, ax = plt.subplots()
    else:
        fig = ax.figure

    # Compute limits of plot
    if xlims:
        xmin, xmax = xlims
    else:
        xmin, xmax = rv.ppf(0.000000001), rv.ppf(0.99999)
    xs = np.arange(xmin, xmax)

    # Compute the probability mass function and plot it
    fXs = rv.pmf(xs)
    fXs = np.where(fXs == 0, np.nan, fXs)  # set zero fXs to np.nan
    ax.stem(fXs, basefmt=" ", label=label)
    ax.set_xticks(xs)
    ax.set_xlabel(rv_name.lower())
    ax.set_ylabel(f"$f_{{{rv_name}}}$")
    if ylims:
        ax.set_ylim(*ylims)
    if label:
        ax.legend()

    if title and title.lower() == "auto":
        title = "Probability mass function of the random variable " + rv.dist.name + str(rv.args)
    if title:
        ax.set_title(title, y=0, pad=-30)

    # return the axes
    return ax


def plot_cdf(rv, xlims=None, ylims=None, rv_name="X", ax=None, title=None, label=None):
    """
    Plot the CDF of the random variable `rv` (discrete or continuous) over the `xlims`.
    """
    # Setup figure and axes
    if ax is None:
        fig, ax = plt.subplots()
    else:
        fig = ax.figure

    # Compute limits of plot
    if xlims:
        xmin, xmax = xlims
    else:
        xmin, xmax = rv.ppf(0.000000001), rv.ppf(0.99999)
    xs = np.linspace(xmin, xmax, 1000)

    # Compute the CDF and plot it
    FXs = rv.cdf(xs)
    sns.lineplot(x=xs, y=FXs, ax=ax)

    # Set plot attributes
    ax.set_xlabel(rv_name.lower())
    ax.set_ylabel(f"$F_{{{rv_name}}}$")
    if ylims:
        ax.set_ylim(*ylims)
    if label:
        ax.legend()
    if title and title.lower() == "auto":
        title = "Cumulative distribution function of the random variable " + rv.dist.name + str(rv.args)
    if title:
        ax.set_title(title, y=0, pad=-30)

    # return the axes
    return ax



# Continuous random variables
################################################################################

def plot_pdf(rv, xlims=None, ylims=None, rv_name="X", ax=None, title=None, **kwargs):
    """
    Plot the pdf of the continuous random variable `rv` over the `xlims`.
    """
    # Setup figure and axes
    if ax is None:
        fig, ax = plt.subplots()
    else:
        fig = ax.figure

    # Compute limits of plot
    if xlims:
        xmin, xmax = xlims
    else:
        xmin, xmax = rv.ppf(0.000000001), rv.ppf(0.99999)
    xs = np.linspace(xmin, xmax, 1000)

    # Compute the probability density function and plot it
    fXs = rv.pdf(xs)
    sns.lineplot(x=xs, y=fXs, ax=ax, **kwargs)
    ax.set_xlabel(rv_name.lower())
    ax.set_ylabel(f"$f_{{{rv_name}}}$")
    if ylims:
        ax.set_ylim(*ylims)

    if title and title.lower() == "auto":
        title = "Probability density function of the random variable " + rv.dist.name + str(rv.args)
    if title:
        ax.set_title(title, y=0, pad=-30)

    # return the axes
    return ax





# Diagnostic plots (used in Section 2.7 Random variable generation)
################################################################################
# The function qq_plot tries to imitate the behaviour of the function `qqplot`
# defined in `statsmodels.graphics.api`. Usage: `qqplot(data, dist=norm(0,1), line='q')`. See:
# https://github.com/statsmodels/statsmodels/blob/main/statsmodels/graphics/gofplots.py#L912-L919
#
# TODO: figure out how to plot all of data correctly: currently missing first and last data point

def qq_plot(data, dist, ax=None, xlims=None, filename=None):
    # Setup figure and axes
    if ax is None:
        fig, ax = plt.subplots()
    else:
        fig = ax.figure

    # Add the Q-Q scatter plot
    qs = np.linspace(0, 1, len(data)+1)
    xs = dist.ppf(qs)
    ys = np.quantile(data, qs)
    sns.scatterplot(x=xs, y=ys, ax=ax, alpha=0.2)

    # Compute the parameters m and b for the diagonal
    xq25, xq75 = dist.ppf([0.25, 0.75])
    yq25, yq75 = np.quantile(data, [0.25,0.75])
    m = (yq75-yq25)/(xq75-xq25)
    b = yq25 - m * xq25
    # add the line  y = m*x+b  to the plot
    linexs = np.linspace(min(xs[1:]),max(xs[:-1]))
    lineys = m*linexs + b
    sns.lineplot(x=linexs, y=lineys, ax=ax, color="r")

    # Handle keyword arguments
    if xlims:
        ax.set_xlim(xlims)
    if filename:
        basename = filename.replace('.pdf','').replace('.png','')
        fig.tight_layout()
        fig.savefig(basename + '.pdf', dpi=300, bbox_inches="tight", pad_inches=0.02)
        fig.savefig(basename + '.png', dpi=300, bbox_inches="tight", pad_inches=0.02)

    return ax



import numpy as np

# SAMPLING DISTRIBUTIONS
################################################################################

# This function is used in Chapter 2 with argument `statfunc` because the concept
# of estimator has not been introduced yet, but it is identical to the function below.

def gen_sampling_dist(rv, statfunc, n, N=1000):
    """
    Simulate `N` samples of size `n` from the random variable `rv` to
    generate the sampling distribution of the statistic `statfunc`.
    """
    stats = []
    for i in range(0, N):
        sample = rv.rvs(n)
        stat = statfunc(sample)
        stats.append(stat)
    return stats



# This function is identical to the one above, but uses the argument `estfunc`.
# It is the main function used in Chapter 3 and later.

def gen_sampling_dist(rv, estfunc, n, N=10000):
    """
    Simulate `N` samples of size `n` from the random variable `rv` to
    generate the sampling distribution of the estimator `estfunc`.
    """
    estimates = []
    for i in range(0, N):
        sample = rv.rvs(n)
        estimate = estfunc(sample)
        estimates.append(estimate)
    return estimates




# BOOTSTRAP
################################################################################

def gen_boot_dist(sample, estfunc, B=5000):
    """
    Generate estimates from the sampling distribution of the estimator `estfunc`
    based on `B` bootstrap samples (sampling with replacement) from `sample`.
    """
    n = len(sample)
    bestimates = []
    for i in range(0, B):
        bsample = np.random.choice(sample, n, replace=True)
        bestimate = estfunc(bsample)
        bestimates.append(bestimate)
    return bestimates




from collections import defaultdict
import os

import numpy as np
import pandas as pd
from scipy.stats import bootstrap

from .confidence_intervals import ci_mean
from .confidence_intervals import ci_var
from .estimators import var


def boot_ci(sample, estfunc, alpha=0.1, method=None, B=5000):
    """
    An adaptor for calling the function `scipy.stats.bootstrap` without the need
    to specify all all the optional keyword arguments.
    """
    res = bootstrap([sample],
                    statistic=estfunc,
                    confidence_level=1-alpha,
                    n_resamples=B,
                    vectorized=False,
                    method=method)
    return [res.confidence_interval.low,
            res.confidence_interval.high]



def simulate_ci_props(pops, methods=["a", "percentile", "bca"], ns=[20,40],
                      param="mean", alpha=0.1, N=1000, B=5000, seed=42):
    """
    Runs a simulation of confidence intervals for `param` using `methods`
    for sample sizes `ns` from populations `pops` (dict label: model).
    Simulation parameters:
        - pops          # populations
        - methods       = ["a", "percentile", "bca"]
        - ns = [20,40]  # sample sizes
        - param         # population parameter
        - alpha = 0.1   # target error level
        - N = 1000      # number of simulations
        - B = 5000      # number of bootstrap samples
    """
    assert param in ["mean", "var"]

    # check if cached simulation data exists
    filename = "simulate_ci_props_" + param + "__ns_" + "_".join(map(str,ns)) \
                + "__alpha_" + str(alpha) + "__seed_" + str(seed) + ".csv"
    filepath = os.path.join("simdata", filename)
    if os.path.exists(filepath):  # load cached results
        print("loaded cached results from ", filepath)
        results = pd.read_csv(filepath, header=[0,1], index_col=[0,1])
        return results

    # simulation data structures
    rowsindex = pd.MultiIndex.from_product((pops.keys(),ns), names=["population", "n"])
    colindex = pd.MultiIndex.from_product((["wbar", "cov"], methods), names=["property", "method"])
    widthscolindex = pd.MultiIndex.from_product((methods,ns), names=["method","n"])
    results = pd.DataFrame(index=rowsindex, columns=colindex)

    # run simulation
    np.random.seed(seed)
    print("Starting simulation for confidence intervals of population {param} :::::::::::::")
    for pop in pops.keys():
        print(f"Evaluating rv{pop} ...")
        rv = pops[pop]
        if param == "mean":
            pop_param = rv.mean()
        elif param == "var":
            pop_param = rv.var()
        counts = defaultdict(int)  # keys are tuples (method,n)
        widths = pd.DataFrame(index=range(0,N), columns=widthscolindex)
        for n in ns:
            print(f"  - running simulation with {n=} ...")
            for j in range(0, N):
                sample = rv.rvs(n)
                for method in methods:
                    if method == "a":
                        if param == "mean":
                            ci = ci_mean(sample, alpha=alpha, method="a")
                        elif param == "var":
                            ci = ci_var(sample, alpha=alpha, method="a")
                    else:
                        if param == "mean":
                            ci = boot_ci(sample, estfunc=np.mean, alpha=alpha, method=method, B=B)
                        elif param == "var":
                            ci = boot_ci(sample, estfunc=var, alpha=alpha, method=method, B=B)
                    # evaluate confidence interval parameters
                    if ci[0] <= pop_param <= ci[1]:
                        counts[(method,n)] += 1  # success
                    # width
                    widths.loc[j,(method,n)] = ci[1] - ci[0]
        for method in methods:
            for n in ns:
                results.loc[(pop, n), ("cov", method)] = counts[(method, n)] / N
                results.loc[(pop, n), ("wbar", method)] = widths.mean()[method, n]

    results.to_csv(filepath)
    print("Saved file to " + filepath)
    return results




import numpy as np
from scipy.optimize import fmin

def hpdi_from_grid(params, probs, hdi_prob=0.9):
    """
    Compute the highest probability density interval from
    the grid approximation probabilities `probs` evaluated at `params`.
    """
    assert len(params) == len(probs)
    cdfprobs = np.cumsum(probs)

    def ppf(prob):
        "Inverse CDF function calculated from `params` grid and `prob`."
        idx = cdfprobs.searchsorted(prob)
        return params[idx]

    def width(startprob):
        "Calculate the width of the `hdi_prob` interval starting at `startprob`."
        return ppf(startprob + hdi_prob) - ppf(startprob)
    
    idx_left_stop = cdfprobs.searchsorted(1 - hdi_prob)
    widths = []
    for idx_probstart in range(0, idx_left_stop):
        probstart = cdfprobs[idx_probstart]
        widths.append(width(probstart))
    
    idx_left = np.argmin(widths)
    hdi_left = params[idx_left]
    probstart = cdfprobs[idx_left]    
    idx_right = cdfprobs.searchsorted(probstart + hdi_prob)
    hdi_right = params[idx_right]
    return [hdi_left, hdi_right]



def hpdi_from_rv(rv, hdi_prob=0.9):
    probstart0 = (1 - hdi_prob) / 2
    def width(probstart):
        return rv.ppf(probstart + hdi_prob) - rv.ppf(probstart)
    min_probstart = fmin(width, x0=probstart0, xtol=1e-9, disp=False)[0]
    hdi_left = rv.ppf(min_probstart)
    hdi_right = rv.ppf(min_probstart + hdi_prob)
    return [hdi_left, hdi_right]



def hpdi_from_samples(samples, hdi_prob=0.9):
    samples = np.sort(samples)
    n = len(samples)
    idx_width = int(np.floor(hdi_prob*n))
    n_intervals = n - idx_width
    left_endpoints = samples[0:n_intervals]
    right_endpoints = samples[idx_width:n]
    widths = right_endpoints - left_endpoints
    idx_left = np.argmin(widths)
    hdi_left = samples[idx_left]
    hdi_right = samples[idx_left + idx_width]
    return [hdi_left, hdi_right]




import math

import arviz
import numpy as np
from scipy.stats import norm
from ministats.probs import MixtureModel
import pytest

# SUT
from ministats.hpdi import hpdi_from_grid
from ministats.hpdi import hpdi_from_rv
from ministats.hpdi import hpdi_from_samples


@pytest.fixture
def rvX1():
    rvX1 = MixtureModel([norm(0,1), norm(3,2)], weights=[0.4,0.6])
    return rvX1

@pytest.fixture
def samplesX1(rvX1, seed=44):
    nsamples = 100_000
    np.random.seed(seed)
    samples = rvX1.rvs(nsamples)
    return samples

@pytest.fixture
def gridX1(rvX1):
    ngrid = 10_000
    lims = [-5, 15]
    params = np.linspace(*lims, ngrid)
    probs = rvX1.pdf(params)
    probs /= np.sum(probs)
    return params, probs


def test_hpdi_from_grid(gridX1, samplesX1):
    arviz_ci = arviz.hdi(samplesX1, hdi_prob=0.9)
    params, probs = gridX1
    ci = hpdi_from_grid(params, probs, hdi_prob=0.9)
    print(arviz_ci - ci)
    assert math.isclose(arviz_ci[0], ci[0], abs_tol=0.03)
    assert math.isclose(arviz_ci[1], ci[1], abs_tol=0.03)

def test_hpdi_from_samples(samplesX1):
    arviz_ci = arviz.hdi(samplesX1, hdi_prob=0.9)
    ci = hpdi_from_samples(samplesX1, hdi_prob=0.9)
    assert math.isclose(arviz_ci[0], ci[0])
    assert math.isclose(arviz_ci[1], ci[1])


def test_hpdi_from_rv(rvX1, samplesX1):
    arviz_ci = arviz.hdi(samplesX1, hdi_prob=0.9)
    ci = hpdi_from_rv(rvX1, hdi_prob=0.9)
    print(arviz_ci - ci)
    assert math.isclose(arviz_ci[0], ci[0], abs_tol=0.01)
    assert math.isclose(arviz_ci[1], ci[1], abs_tol=0.01)
